import contextlib
import logging
import signal
import time
from multiprocessing import Event, Process

import django
from django import db
from django.apps import apps

from django_kafka import kafka
from django_kafka.exceptions import DjangoKafkaError

logger = logging.getLogger(__name__)


class KafkaConsumeRunner:
    def __init__(self, consumers: list[str]):
        self.consumers = consumers
        self.processes: list[Process] = []
        self.stop_event: Event = Event()

    def start(self):
        signal.signal(signal.SIGINT, self._soft_shutdown)  # Ctrl+C or kill -SIGINT
        signal.signal(signal.SIGTERM, self._soft_shutdown)  # kill <pid>

        db.connections.close_all()  # close before spawning new processes

        try:
            # spawn one process per consumer
            for key in self.consumers:
                worker = ConsumerWorker(key, self.stop_event)
                process = Process(target=worker.start, name=f"consumer-{key}")
                process.start()
                self.processes.append(process)

            # waiting loop
            while any(p.exitcode is None for p in self.processes):
                if any(p.exitcode not in (None, 0) for p in self.processes):
                    # shut down if any has failed.
                    self._soft_shutdown(None, None)
                    raise DjangoKafkaError(
                        "The consumer runner process exited unexpectedly.",
                    )
                time.sleep(0.2)

        finally:
            for process in self.processes:
                # timeout is to avoid infinite wait, if anything goes wrong
                process.join(timeout=1.0)
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            signal.signal(signal.SIGTERM, signal.SIG_DFL)

    def _soft_shutdown(self, signum, frame):
        logger.info("Soft shutdown - waiting for consumers to finish")
        if signum == signal.SIGINT:
            logger.warning(
                "Hitting Ctrl+C again will kill running processes immediately!",
            )
        self.stop_event.set()
        if signal.getsignal(signal.SIGINT) is not self._hard_shutdown:
            signal.signal(signal.SIGINT, self._hard_shutdown)

    def _hard_shutdown(self, signum, frame):
        logger.info("Hard shutdown - immediately stopping consumers")

        for process in self.processes:
            if process.is_alive():
                with contextlib.suppress(Exception):
                    process.terminate()

        # little waiting loop, to cover delays in terminations
        end = time.time() + 2  # loop for 2 seconds maximum
        while time.time() < end and any(p.is_alive() for p in self.processes):
            time.sleep(0.05)

        # if someone is still alive - hard kill
        for process in self.processes:
            if process.is_alive():
                process.kill()
        # after hard shutdown, make further SIGINT reset to default
        signal.signal(signal.SIGINT, signal.SIG_DFL)


class ConsumerWorker:
    """
    Consumer worker started by subprocesses, so all internal state must be pickleable.
    """

    def __init__(self, consumer_key: str, stop_event: Event):
        self.consumer_key = consumer_key
        self.stop_event = stop_event

    def start(self):
        # Only parent handles Ctrl+C (SIGINT)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # Ensure SIGTERM in the child uses the default action (terminate)
        # to avoids inheriting the parent's _shutdown handler
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        if not apps.ready:
            django.setup()  # for spawn or forkserver process start methods

        try:
            kafka.consumers[self.consumer_key]().start(self.stop_event)
        except Exception:
            logger.exception("Consumer %s crashed", self.consumer_key)
            raise



from django.contrib import admin

from django_kafka.models import KeyOffsetTracker, WaitingMessage


@admin.register(KeyOffsetTracker)
class KeyOffsetTrackerAdmin(admin.ModelAdmin):
    list_display = ("topic", "key", "offset")
    search_fields = ("topic", "key")


@admin.register(WaitingMessage)
class WaitingMessageAdmin(admin.ModelAdmin):
    list_display = (
        "topic",
        "partition",
        "offset",
        "relation_model_key",
        "relation_id_field",
        "relation_id_value",
        "status",
    )
    search_fields = (
        "topic",
        "partition",
        "offset",
        "relation_model_key",
        "relation_id_field",
        "relation_id_value",
        "key",
        "status",
    )



import logging
from collections.abc import Callable
from contextlib import ContextDecorator
from contextvars import ContextVar
from pydoc import locate

from confluent_kafka import Producer as ConfluentProducer

from django_kafka.conf import settings

logger = logging.getLogger(__name__)


class Producer:
    """
    Available settings of the producers (P) and consumers (C):
        https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
    Producer configs
        https://kafka.apache.org/documentation/#producerconfigs
    Kafka Client Configuration
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafka-client-configuration
    confluent_kafka.Producer API
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-producer
    """

    config: dict

    default_logger = logger
    default_error_handler = settings.ERROR_HANDLER

    def __init__(self, config: dict | None = None, **kwargs):
        kwargs.setdefault("logger", self.default_logger)
        kwargs.setdefault("error_cb", locate(self.default_error_handler)())

        self._producer = ConfluentProducer(
            {
                "client.id": settings.CLIENT_ID,
                **settings.GLOBAL_CONFIG,
                **settings.PRODUCER_CONFIG,
                **getattr(self, "config", {}),
                **(config or {}),
            },
            **kwargs,
        )

    def produce(self, name, *args, **kwargs):
        if not Suppression.active(name):
            self._producer.produce(name, *args, **kwargs)

    def __getattr__(self, name):
        """
        proxy producer methods.
        """
        if name not in {"config"}:
            # For cases when `Producer.config` is not set and
            #  `getattr(self, "config", {})` is called on `__init__`,
            #  the initialization will fail because `_consumer` is not yet set.
            return getattr(self._producer, name)
        raise AttributeError(f"'{self.__class__.__name__}' has no attribute 'name'")


class Suppression(ContextDecorator):
    """context manager to help suppress producing messages to desired Kafka topics"""

    _var = ContextVar(f"{__name__}.suppression")

    @classmethod
    def active(cls, topic: str):
        """returns if suppression is enabled for the given topic"""
        try:
            topics = cls._var.get()
        except LookupError:
            # context var is not yet set, defaulting to empty list
            topics = []

        # topics will be None when suppress() is initialized without topics provided
        if topics is None:
            return True  # all topics

        return topic in topics

    def __init__(self, topics: list[str] | None = None, deactivate=False):
        try:
            topics_in_context = self._var.get()
        except LookupError:
            topics_in_context = []

        if deactivate:
            self.topics = []

        elif topics is None or topics_in_context is None:
            self.topics = None  # indicates all topics

        elif isinstance(topics, list):
            self.topics = topics_in_context + topics

        else:
            raise ValueError(f"invalid producer suppression setting {topics}")

    def __enter__(self):
        self.token = self._var.set(self.topics)
        return self

    def __exit__(self, *args, **kwargs):
        self._var.reset(self.token)


def suppress(topics: Callable | list[str] | None = None):
    if callable(topics):
        return Suppression()(topics)
    return Suppression(topics)


def unsuppress(fn: Callable | None = None):
    if fn:
        return Suppression(deactivate=True)(fn)
    return Suppression(deactivate=True)
